FUNCTION FINDGAIN, flat1 , bias1 , flat2 , bias2 , RON=ron , CLIP=clip
;+
; NAME:
;
;       FINDGAIN
;
; PURPOSE:
;
;       Returns the gain and read out noise, given a pair of BIAS exposures
;	and a pair of equally illuminated flats.
;
; CALLING SEQUENCE:
;
;       gain	 = FINDGAIN( flat1 , bias1  , flat2 , bias2 , [ RON=ron $
;                , CLIP=clip ] )
;
; INPUTS:
;
;       FLAT1		1.st flat field exposure. Can be the name of a 
;			two or three dimensional FITS file or a 
;			two or three dimensional data array.
;
;       BIAS1		1.st bias exposure. Can be the name of a 
;			two or three dimensional FITS file or a 
;			two or three dimensional data array.
;
;       FLAT2		2.nd flat field exposure. Can be the name of a 
;			two or three dimensional FITS file or a 
;			two or three dimensional data array.
;			If FLAT1 is three dimensional, FLAT2 may be omitted.
;
;       BIAS2		2.nd bias exposure. Can be the name of a 
;			two or three dimensional FITS file or a 
;			two or three dimensional data array.
;			If BIAS1 is three dimensional, BIAS2 may be omitted.
;
; OPTIONAL KEYWORDS
;
;	RON		Returned as read out noise in electorns.
;
;	CLIP		If set, sigma clipping at CLIP standard deviations
;			is performed.
;
;	NITER		Number of clipping iterations, when 
;			CLIP keyword is set. Default value is 2.
;
; OUTPUTS:
;
;       GAIN :		The estimated gain.
;
; RESTRICTIONS
;
; 	uses non-standard procedure IDL_TYPE
;
; MODIFICATION HISTORY:
;
;       WRITTEN, Michael Andersen CUOBS, December, 1995
;	BUG-FIX: CLIP parameter now specifies sigma clipping,
;	in stead of turning on a fixed 3-sigma clipping.
;-

  ON_ERROR, 2

; read data, if input is FITS files.

  IF ( IDL_TYPE( flat1 ) EQ 7 ) THEN BEGIN
    f1file	= flat1
    b1file	= bias1
    flat1	= READFITS( f1file , hflat1 , /SILENT )
    bias1	= READFITS( b1file , hbias1 , /SILENT )
  ENDIF

  ;Undg crash ved totalt saturerede flats
  sfdiff=0.
      
  IF ( N_PARAMS() EQ 2 AND (SIZE( flat1 ))(0) NE 3 ) THEN $
  MESSAGE, "Must specify two flats and two bias frames or 3D data"

  IF ( N_PARAMS() GT 2 AND IDL_TYPE( flat2 ) EQ 7 ) THEN BEGIN
    f2file	= flat2
    b2file	= bias2
    flat2	= READFITS( f2file , hflat2 , /SILENT )
    bias2	= READFITS( b2file , hbias2 , /SILENT )
  ENDIF

; Calculate gain and ron

  IF ( N_PARAMS() EQ 4 ) THEN BEGIN

    fdiff	= flat2 - flat1
    bdiff	= bias2 - bias1
    avgf1	= MEAN( flat1 )
    avgf2	= MEAN( flat2 )
    avgb1	= MEAN( bias1 )
    avgb2	= MEAN( bias2 )

    if (avgf1 le 0) then begin
      print,"Flat 1: Average value not positive!"
      stop
    endif

    if (avgf2 le 0) then begin
      print,"Flat 2: Average value not positive!"
      stop
    endif

    if (avgb1 le 0) then begin
      print,"Bias 1: Average value not positive!"
      stop
    endif

    if (avgb2 le 0) then begin
      print,"Bias 2: Average value not positive!"
      stop
    endif


    
  ENDIF ELSE BEGIN

    naxis3	= (SIZE( flat1 ))(3) / 2 * 2
    i		= 2 * LINDGEN( naxis3 / 2 )
    j		= i + 1 
    fdiff	= flat1(*,*,j) - flat1(*,*,i)
    bdiff	= bias1(*,*,j) - bias1(*,*,i)
    avgf1	= MEAN( flat1(*,*,i) )
    avgf2	= MEAN( flat1(*,*,j) )
    avgb1	= MEAN( bias1(*,*,i) )
    avgb2	= MEAN( bias1(*,*,j) )


  ENDELSE

  IF ( (SIZE(fdiff))(0) EQ 2 ) THEN BEGIN
    IF ( NOT KEYWORD_SET( CLIP ) ) THEN BEGIN
      sfdiff	= STDDEV( fdiff )
      sbdiff	= STDDEV( bdiff )
    ENDIF ELSE BEGIN
      afdiff	= SIGCLIP_AVG( fdiff , CLIP , CLIP , RMS = sfdiff )
      abdiff	= SIGCLIP_AVG( bdiff , CLIP , CLIP , RMS = sbdiff )
      ;stop
    ENDELSE
  ENDIF ELSE BEGIN
    naxis3	= (SIZE(fdiff))(3)
    sfdiff	= FLTARR( naxis3 )
    sbdiff	= FLTARR( naxis3 )
    FOR i = 0 , naxis3-1 DO BEGIN
      IF ( NOT KEYWORD_SET( CLIP ) ) THEN BEGIN
        sfdiff(i)	= STDDEV( fdiff(*,*,i) )
        sbdiff(i)	= STDDEV( bdiff(*,*,i) )
      ENDIF ELSE BEGIN
        afdiff	= SIGCLIP_AVG( fdiff(*,*,i) , CLIP , CLIP , RMS = sf )
        abdiff	= SIGCLIP_AVG( bdiff(*,*,i) , CLIP , CLIP , RMS = sb )
	sfdiff(i) = sf
	sbdiff(i) = sb
      ENDELSE
    ENDFOR
  ENDELSE

  gain		= ( (avgf1+avgf2) - (avgb1+avgb2) ) / ( sfdiff^2 - sbdiff^2 )
  ron		= gain * sbdiff / SQRT( 2 )

  RETURN, gain

END
